/*******************************************************************************
 * Copyright (C) 2024 Juan Neyra
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
#pragma once

#ifndef SC_INCLUDE_DYNAMIC_PROCESSES
#define SC_INCLUDE_DYNAMIC_PROCESSES
#endif

#include <iostream>
#include <string>
#include <vector>
#include <map>

#include "systemc.h"
#include "tlm.h"

#include "ratatoskrUtils/utils/Structures.h"
#include "utils/GlobalResourcesCS.h"
#include "ratatoskrUtils/utils/GlobalReport.h"

#include "router/router.h"
#include "router/router_cs.h"

using namespace tlm;

class TlmNoc : public sc_module {
public:
    typedef tlm_utils::simple_initiator_socket_tagged<TlmRouter> tlm_init_socket;
    typedef tlm_utils::simple_target_socket_tagged<TlmRouter> tlm_targ_socket;

    SC_HAS_PROCESS(TlmNoc);

    explicit TlmNoc(sc_module_name, string);

    ~TlmNoc() override;

    std::vector<tlm_init_socket*> api_initiators;
    std::vector<tlm_targ_socket*> api_targets;
    uint8_t max_pos[3];
    std::map<int, int> mapNodeRouter;
    TlmRouter* getRouter(int id);
    TlmRouter* getRouterNodeId(int node_id);

private:
    std::vector<TlmRouter*> routers;
    uint8_t num_rout_x;
    uint8_t num_rout_y;
    uint8_t num_rout_z;
    std::vector<tlm_init_socket*> unbounded_initiators;
    std::vector<tlm_targ_socket*> unbounded_targets;

    GlobalResourcesCS& globalResources = GlobalResourcesCS::getInstance();
    GlobalResources& globalResources_legacy = GlobalResources::getInstance();

    /** 
     * Initialize global resources
     */
    void initializeGlobalResources(string config_folder);

    /** 
     * Initialize variable in noc
     */
    void initNoc();


    /** 
     * Connects two routers or a router and a core (network interface)
     * 
     * @param node1 first node to connect
     * @param node2 second node to connect
     * @param link  link of first node that will be connected
     */
    void connectRouters(Node node1, Node node2, int link);

    /** 
     * Connects unbounded links to dummy initiators and targets
     */
    void connectUnbounded();

    /** 
     * Connects all routers using GlobalResources object
     */
    void connectNoc();
    

    /** Log info
     * @param msg  log message
     */
    void log_info(string msg);

    /** Log warning
     * @param msg  log message
     */
    void log_warn(string msg);

    /** Log error
     * @param msg  log message
     */
    void log_error(string msg);

    /** Log a connection between two routers
     * @param router1  name of router 1
     * @param router2  name of router 2
     * @param link1    name of connected link in router 1
     * @param link2    name of connected link in router 2
     */
    void log_conn(string router1, string router2,
                        string link1, string link2);
};
