/*******************************************************************************
 * Copyright (C) 2024 Juan Neyra
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
#include "router.h"

#include <cassert>
#include <array>
#include "ratatoskrUtils/utils/Structures.h"

/* Constructor */
TlmRouter::TlmRouter(sc_module_name name, uint8_t rout_pos[3], 
                                        uint8_t max_pos[3]):
                sc_module(name), router_name(name),
                init_peq(this, &TlmRouter::init_peq_cb),
                target_peq(this, &TlmRouter::target_peq_cb) {
    sc_report_handler::set_actions(R_LOG, SC_INFO, SC_LOG|SC_DISPLAY);
    copy(rout_pos, rout_pos + 3, this->rout_pos);
    copy(max_pos, max_pos + 3, this->max_pos);
    initialize();
}

TlmRouter::~TlmRouter(){
}

void TlmRouter::initialize(){
    for(int link=0; link<NUM_LINKS; link++){
        init_socket[link] = new rout_init_socket((router_name+"_"+
                            to_string(link)+"_init").c_str());
        (*init_socket[link]).register_nb_transport_bw(this, 
                            &TlmRouter::nb_transport_bw_cb, link);

        target_socket[link] = new rout_targ_socket((router_name+"_"+
                            to_string(link)+"_targ").c_str());
        (*target_socket[link]).register_nb_transport_fw(this, 
                            &TlmRouter::nb_transport_fw_cb, link);

        valid_links[link] = false;
        resp_in_progress[link] = false;
        //nxt_resp_pend[link] = 0;
        //nxt_send_data_pend[link] = 0;
        curr_req[link] = 0;
        //send_data_in_prog_dest[link] = Direction::invalid;
        credit_counter[link] = NUM_CREDITS;
    }
}

/******************* COMMON FUNCTIONS ********************/
tlm_gp* TlmRouter::build_transaction(tlm_gp& trans, int dest_link){
    tlm::tlm_generic_payload* new_trans = m_mm.allocate();
    new_trans->acquire();
    new_trans->set_command(tlm::TLM_WRITE_COMMAND);
    new_trans->set_address(trans.get_address());
    new_trans->set_data_ptr(trans.get_data_ptr());
    new_trans->set_data_length(trans.get_data_length());
    new_trans->set_streaming_width(trans.get_streaming_width());
    new_trans->set_byte_enable_ptr(trans.get_byte_enable_ptr());
    new_trans->set_dmi_allowed(false);
    new_trans->set_response_status(tlm::TLM_INCOMPLETE_RESPONSE);
    
    // construct new link extension from old one
    link_extension* ext = new link_extension();
    link_extension* copy_ext;
    trans.get_extension<link_extension>(copy_ext);
    ext->copy_from(*copy_ext);
    ext->link = dest_link; // set direction of connected router
    new_trans->set_extension(ext);

    return new_trans;
}

void TlmRouter::send_begin_req(int link, tlm_gp& trans, int dest_link){
    tlm_gp* new_trans = build_transaction(trans, dest_link);

    // send transaction in socket
    tlm_phase phase = BEGIN_REQ;
    sc_time delay = sc_time(REQ_INIT_DELAY, UNITS_DELAY);

    tlm_sync_enum status;
    status = (*init_socket[dest_link])->nb_transport_fw(
                                *new_trans, phase, delay);
    curr_req[dest_link] = new_trans;
    credit_counter[dest_link]--;
    log_info(link, trans, "credits value updated on link " + DIR::toString(dest_link) +
                          ": " + to_string(credit_counter[dest_link]));
    // react to result
    if(status == TLM_UPDATED) {
        init_peq.notify(*new_trans, phase, delay);
    }
    else if(status == TLM_COMPLETED) {
        log_warn(link,trans,"Request completed prematurely");
        curr_req[dest_link] = 0;
        check_transaction(link, *new_trans);
        credit_counter[dest_link]++;
        log_info(link, trans, "credits value updated on link " + DIR::toString(dest_link) +
                              ": " + to_string(credit_counter[dest_link]));
        new_trans->release();
    }
}


bool TlmRouter::send_data(int link, int dest_link, tlm_gp& trans){
    log_info(link,trans,"Send data to link:"+DIR::toString(dest_link));
    // check credits or that previous message is complete
    if (credit_counter[dest_link] > 0){
        send_begin_req(link, trans, dest_link);
        // send response or add it to queue
        if(resp_in_progress[link]) {
            nxt_resp_pend[link].push(&trans);
        }
        else{ send_begin_response(link, trans); }

        return true;
    }
    else {
        log_warn(link,trans, "Waiting for downstream Router!");
        report.update_throttled_count(router_name);
        return false;
    }
}

void TlmRouter::check_transaction(int link, tlm_gp& trans) {
    if(trans.is_response_error()) {
        char txt[100];
        sprintf(txt, "Transaction returned with error, response status = %s",
                    trans.get_response_string().c_str());
        log_warn(link,trans,txt);
    }

    // Log completed routing
    tlm_command cmd = trans.get_command();
    sc_dt::uint64    adr = trans.get_address();
    int*             ptr = reinterpret_cast<int*>(trans.get_data_ptr());
    std::stringstream stream;
    stream << hex << adr << " check, cmd=" << (cmd ? 'W' : 'R')
                         << ", data=" << hex << *ptr;
    log_info(link,trans, stream.str());
    trans.release();
}

int TlmRouter::get_link_from_extension(tlm_gp& trans){
    link_extension* extension;
    trans.get_extension<link_extension>(extension);
    return extension->link;
}

string TlmRouter::get_id_string_from_extension(tlm_gp& trans){
    link_extension* ext;
    trans.get_extension<link_extension>(ext);
    string str_pck_id = ext->packet_id > 0 ? to_string(ext->packet_id):"none";
    string str_flt_id = ext->flit_id > 0 ? to_string(ext->flit_id):"none";
    return "packet " + to_string(ext->packet_id) + 
           ":flit " + to_string(ext->flit_id);
}

/******************* INIT SOCKET FUNCTIONS ********************/
tlm_sync_enum TlmRouter::nb_transport_bw_cb(int id, tlm_gp& trans, 
                                tlm_phase& phase, sc_time& delay) {
    log_info(id, trans, "Backward transport callback start");
    init_peq.notify(trans, phase, delay);
    return TLM_ACCEPTED;
}

void TlmRouter::init_peq_cb(tlm_gp& trans, const tlm_phase& phase){
    int link = get_link_from_extension(trans);
    log_info(link,trans, "Initiator PEQ callback start");
    log_info(link,trans, "Phase "+string(phase.get_name())+" received");
    if (phase == END_REQ || (&trans == curr_req[link] && phase == BEGIN_RESP)){
        curr_req[link] = 0;
        /*credit_counter[link]++;
        log_info(link, trans, "credits value updated on link " + DIR::toString(link) +
                ": " + to_string(credit_counter[link]));*/
        log_info(link, trans, "Issuing next request");
        start_pend_req(link, trans);
    }
    else if (phase == BEGIN_REQ || phase == END_RESP){
        log_error(link,trans,"Illegal transaction phase received by initiator");
    }

    if (phase == BEGIN_RESP) {
        credit_counter[link]++;
        log_info(link, trans, "credits value updated on link " + DIR::toString(link) +
                ": " + to_string(credit_counter[link]));
        send_end_response(link, trans);
    }
}


void TlmRouter::start_pend_req(int rel_link, tlm_gp& trans){
    // find pending link
    for(uint8_t dir=0; dir<Direction::num_dirs; dir++){
        if( !nxt_send_data_pend[dir].empty() ){
            // check pending link valid and send data
            Direction rel_dest = send_data_in_prog_dest[dir].front();
            if(rel_dest == rel_link){
                tlm_gp* nxt_trans = nxt_send_data_pend[dir].front();
                bool data_sent = send_data(dir, rel_link, *nxt_trans);
                if (!data_sent){
                    log_warn(dir, trans, "Credit counter was freed, but data can't be sent");
                }
                else{
                    send_data_in_prog_dest[dir].pop();
                    nxt_send_data_pend[dir].pop();
                }
                break;
            }
        }
    }
}

void TlmRouter::send_end_response(int link, tlm_gp& trans){
    log_info(link,trans, "Send end response start");
    tlm_phase fw_phase = END_RESP;
    sc_time delay = sc_time(RESP_END_DELAY, UNITS_DELAY);
    (*init_socket[link])->nb_transport_fw(trans, fw_phase, delay);
}


/******************* TARGET SOCKET FUNCTIONS ********************/
tlm_sync_enum TlmRouter::nb_transport_fw_cb(int id,
            tlm_gp& trans, tlm_phase& phase, sc_time& delay){
    log_info(id, trans, "Forward transport callback start");
    unsigned int     len = trans.get_data_length();
    unsigned char*   byt = trans.get_byte_enable_ptr();
    unsigned int     wid = trans.get_streaming_width();
    if (byt != 0) {
        trans.set_response_status(TLM_BYTE_ENABLE_ERROR_RESPONSE);
        return TLM_COMPLETED;
    }
    if (len > 4 || wid < len) {
        trans.set_response_status(TLM_BURST_ERROR_RESPONSE);
        return TLM_COMPLETED;
    }
    target_peq.notify(trans, phase, delay);
    return TLM_ACCEPTED;
}

void TlmRouter::target_peq_cb(tlm_gp& trans, const tlm_phase& phase){
    int link = DIR::getOppositeDir(get_link_from_extension(trans));
    log_info(link,trans, "Target PEQ callback start");
    log_info(link,trans, "Phase "+string(phase.get_name())+" received");
    switch (phase) {
        case BEGIN_REQ:
            trans.acquire();
            send_end_req(link, trans);
            break;
        case END_RESP:
            if (!resp_in_progress[link]){
                log_warn(link,trans,
                    "Illegal transaction phase END_RESP received by target");
            }
            check_transaction(link, trans);
            // Target itself is now clear to issue the next BEGIN_RESP
            resp_in_progress[link] = false;
            if ( !nxt_resp_pend[link].empty() ){
                log_info(link, trans, "Issuing next response");
                send_begin_response(link, *nxt_resp_pend[link].front());
                nxt_resp_pend[link].pop();
            }
            break;
        case END_REQ:
        case BEGIN_RESP:
            log_error(link,trans,"Illegal transaction phase received by target");
            break;
        default:
            report.update_packet_count(router_name);
            if(phase == INTERNAL_PROC_PHASE){
                switching(link, trans);
            }
            break;
    }
}

tlm_sync_enum TlmRouter::send_end_req(int link, tlm_gp& trans){
    log_info(link, trans, "Send end request start");
    // Queue the acceptance and the response with the appropriate latency
    sc_time delay = sc_time(REQ_END_DELAY, UNITS_DELAY);
    tlm_phase phase = END_REQ;
    tlm_sync_enum status = (*target_socket[link])->nb_transport_bw(
                                        trans, phase, delay);
    if (status == TLM_COMPLETED) {
        log_warn(link,trans,"Request completed, no response to be send");
        trans.release();
        return status;
    }

    // Queue internal event to mark beginning of response
    delay = delay + sc_time(INTERN_PROC_DELAY, UNITS_DELAY); // Latency
    target_peq.notify(trans, INTERNAL_PROC_PHASE, delay);

    return status;
}

void TlmRouter::send_begin_response(int link, tlm_gp& trans){
    log_info(link,trans, "Send begin response start");
    resp_in_progress[link] = true;
    tlm_phase phase = BEGIN_RESP;
    sc_time zero_delay = SC_ZERO_TIME;
    tlm_sync_enum status = (*target_socket[link])->nb_transport_bw(
                                trans, phase, zero_delay);

    if (status == TLM_UPDATED){
      target_peq.notify(trans, phase, zero_delay);
    }
    else if (status == TLM_COMPLETED) {
      log_warn(link,trans,"Request completed, no response to be send");
      trans.release();
    }
}

void TlmRouter::switching(int link, tlm_gp& trans){
    log_info(link,trans, "Switching start");
    trans.set_response_status(TLM_OK_RESPONSE);
    // arbitration and send message to next router
    Dir destination = routing(link, trans);
    bool data_sent;
    if (destination == Dir::num_dirs) {
        data_sent = true;
        check_transaction(link, trans);
        // change to fatal?
        log_warn(link,trans, "Routing failed. Message couldn't be delivered");
    }
    else{
        data_sent = send_data(link, destination, trans);
    }
    // validate pending data sent
    if (!data_sent) {
        nxt_send_data_pend[link].push(&trans);
        send_data_in_prog_dest[link].push(destination);
    }
}


Dir TlmRouter::xyz_routing(int link, uint8_t dest[3]){
    if (dest[0] < rout_pos[0] && valid_links[Dir::west]){ 
        return Dir::west; 
    }
    else if (dest[0] > rout_pos[0] && valid_links[Dir::east]){ 
        return Dir::east; 
    }
    else if (dest[1] < rout_pos[1] && valid_links[Dir::south]){ 
        return Dir::south;
    }
    else if (dest[1] > rout_pos[1] && valid_links[Dir::north]){ 
        return Dir::north;
    }
    else if (dest[2] < rout_pos[2] && valid_links[Dir::down]){ 
        return Dir::down; 
    }
    else if (dest[2] > rout_pos[2] && valid_links[Dir::up]){ 
        return Dir::up;   
    }
    else if (dest[0] == rout_pos[0] && dest[1] == rout_pos[1] && 
                dest[2] == rout_pos[2]) { 
        return Dir::local;
    }
    else { // no ideal routing found
        return Dir::num_dirs;
    }
}

Dir TlmRouter::routing(int link, tlm_gp& trans){
    int address = trans.get_address();
    uint8_t dest_x = address % max_pos[0];
    uint8_t dest_y = address / max_pos[0];
    uint8_t dest_z = address / (max_pos[0]*max_pos[1]);
    uint8_t dest[3] = {dest_x, dest_y, dest_z};
    return xyz_routing(link, dest);
}


/******************* LOG FUNCTIONS ********************/
void TlmRouter::log_info(uint8_t link, tlm_gp& trans, string msg){
    string str_link = DIR::toString(link);
    string id = get_id_string_from_extension(trans);
    SC_REPORT_INFO(R_LOG, (router_name+":"+str_link+":"+id+":"+msg).c_str());
}

void TlmRouter::log_warn(uint8_t link, tlm_gp& trans, string msg){
    string str_link = DIR::toString(link);
    string id = get_id_string_from_extension(trans);
    SC_REPORT_WARNING(R_LOG, (router_name+":"+str_link+":"+id+":"+msg).c_str());
}

void TlmRouter::log_error(uint8_t link, tlm_gp& trans, string msg){
    string str_link = DIR::toString(link);
    string id = get_id_string_from_extension(trans);
    SC_REPORT_ERROR(R_LOG, (router_name+":"+str_link+":"+id+":"+msg).c_str());
}

void TlmRouter::log_fatal(uint8_t link, tlm_gp& trans, string msg){
    string str_link = DIR::toString(link);
    string id = get_id_string_from_extension(trans);
    SC_REPORT_FATAL(R_LOG, (router_name+":"+str_link+":"+id+":"+msg).c_str());
}
