/*******************************************************************************
 * Copyright (C) 2024 Juan Neyra
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
#pragma once

#ifndef SC_INCLUDE_DYNAMIC_PROCESSES
#define SC_INCLUDE_DYNAMIC_PROCESSES
#endif

#include <systemc>
#include <string>
#include <queue>

#include "tlm.h"
#include "tlm_utils/simple_initiator_socket.h"
#include "tlm_utils/simple_target_socket.h"
#include "tlm_utils/peq_with_cb_and_phase.h"

#include "utils/memory_manager.h"
#include "utils/configuration.h"
#include "utils/utils.h"
#include "utils/noc_logger.h"
#include "utils/report.h"

using namespace sc_core;
using namespace sc_dt;
using namespace std;
using namespace tlm;
using namespace tlm_utils;

#define NUM_LINKS       Direction::num_dirs

DECLARE_EXTENDED_PHASE(INTERNAL_PROC_PHASE);

class TlmRouter : public sc_module{
    public:
        // TLM
        typedef simple_initiator_socket_tagged<TlmRouter> rout_init_socket;
        typedef simple_target_socket_tagged<TlmRouter> rout_targ_socket;
    
        rout_init_socket* init_socket[NUM_LINKS];
        rout_targ_socket* target_socket[NUM_LINKS];
        peq_with_cb_and_phase<TlmRouter> init_peq;
        peq_with_cb_and_phase<TlmRouter> target_peq;
        tlm_gp* curr_req[NUM_LINKS];
        MemoryManager m_mm;
        int credit_counter[NUM_LINKS];
        bool resp_in_progress[NUM_LINKS];
        queue<tlm_gp*> nxt_resp_pend[NUM_LINKS];
        queue<tlm_gp*> nxt_send_data_pend[NUM_LINKS];
        queue<Dir> send_data_in_prog_dest[NUM_LINKS];

        // attributes
        string router_name;
        uint8_t rout_pos[3];
        uint8_t max_pos[3];
        bool valid_links[NUM_LINKS];

        TlmReport& report = TlmReport::getInstance();

        SC_HAS_PROCESS(TlmRouter);
        TlmRouter(sc_module_name name, uint8_t rout_pos[3],
                                      uint8_t max_pos[3]);
        ~TlmRouter();

    protected:
        /**
         * Configure sockets and set initial values of variables
         */
        void initialize();

        /** 
        * Build the transaction tp send. Called by send flit
        *
        * @param trans         TLM generic payload object
        * @param dest_link     destination link to send the data to
        *
        * @return    created transaction
        */
        tlm_gp* build_transaction(tlm_gp& trans, int dest_link);

        /**
         * Send data
         *
         * @param link          active link
         * @param dest_link     destination link to send the data to
         * @param trans         TLM generic payload object
         * @return  true if data was sent, otherwise, false
         */
        bool send_data(int link, int dest_link, tlm_gp& trans);

        /**
         * Checks transaction for errors, if no error found, log
         * transaction and release transaction object. Called when
         * transaction is completed 
         *
         * @param link  active link
         * @param trans TLM generic payload object
         */
        void check_transaction(int link, tlm_gp& trans);

        /**
         * Callback function for non blocking transport backward
         *
         * @param id    active link
         * @param trans TLM generic payload object
         * @param phase TLM current phase
         * @param delay TLM expected delay
         * @return  transaction status
         */
        virtual tlm_sync_enum nb_transport_bw_cb(int id, tlm_gp& trans,
                                tlm_phase& phase, sc_time& delay);

        /**
         * Callback function for non blocking transport forward
         *
         * @param id    active link
         * @param trans TLM generic payload object
         * @param phase TLM current phase
         * @param delay TLM expected delay
         * @return  transaction status
         */
        tlm_sync_enum nb_transport_fw_cb(int id, tlm_gp& trans,
                                tlm_phase& phase, sc_time& delay);

        /**
         * Callback initiator Payload Event Queue (PEQ)
         *
         * @param trans TLM generic payload object
         * @param phase TLM current phase
         */
        virtual void init_peq_cb(tlm_gp& trans, const tlm_phase& phase);

        /**
         * Callback target Payload Event Queue (PEQ)
         *
         * @param trans TLM generic payload object
         * @param phase TLM current phase
         */
        virtual void target_peq_cb(tlm_gp& trans, const tlm_phase& phase);

        /**
         * Begin request
         *
         * @param link      active link
         * @param trans     TLM generic payload object
         * @param dest_link destination link to transmit transaction
         */
        void send_begin_req(int link, tlm_gp& trans, int dest_link);

        /**
         * Send end request
         *
         * @param link active link
         * @param trans TLM generic payload object
         * @return  request status
         */
        tlm::tlm_sync_enum send_end_req(int link, tlm_gp& trans);


        /**
         * Send begin response
         *
         * @param link active link
         * @param trans TLM generic payload object
         */
        void send_begin_response(int link, tlm_gp& trans);

        /**
         * Target itself is now clear to resolve pending data send
         * Executed when request ends
         *
         * @param rel_link link that is now free/released to receive data
         * @param trans TLM generic payload object
         */
        void start_pend_req(int rel_link, tlm_gp& trans);


        /**
         * Send end response
         *
         * @param link  active link
         * @param trans TLM generic payload object
         */
        void send_end_response(int link, tlm_gp& trans);

        /**
         * Process arrived data: finds destination link and transmit it
         *
         * @param link  active link
         * @param trans TLM generic payload object
         */
        void switching(int link, tlm_gp& trans);

        /**
         * xyz routing: x+:east, y+: north, z+: up
         *
         * @param link  active link
         * @param dest  destination
         * @return  direction to route to
         */
        Dir xyz_routing(int link, uint8_t dest[3]);

        /**
         * Routing
         *
         * @param link      active link
         * @param trans     TLM generic payload object
         * @return  direction to route to
         */
        Dir routing(int link, tlm_gp& trans);


        /**
         * Get link from extension
         *
         * @param trans TLM generic payload object
         * @return  link
         */
        int get_link_from_extension(tlm_gp& trans);

        /**
         * Get id from extension as a string (id packet + "_" + id flit)
         *
         * @param trans TLM generic payload object
         * @return  id of packet as string
         */
        string get_id_string_from_extension(tlm_gp& trans);

        /** Log info
         * @param link active link
         * @param msg  log message
         */
        void log_info(uint8_t link, tlm_gp& trans, string msg);

        /** Log warning
         * @param link active link
         * @param msg  log message
         */
        void log_warn(uint8_t link, tlm_gp& trans, string msg);

        /** Log error
         * @param link active link
         * @param msg  log message
         */
        void log_error(uint8_t link, tlm_gp& trans, string msg);

        /** Log fatal
         * @param link active link
         * @param msg  log message
         */
        void log_fatal(uint8_t link, tlm_gp& trans, string msg);
};